import 'dart:math';

import 'package:flutter/material.dart';
import 'package:flutterx_application/src/ui/draggable_bottom_sheet.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

/// Wrapper of [DraggableBottomSheet] which is optimized for matching the screen height
///
/// [initialChildSize] see [DraggableBottomSheet.initialChildSize]
/// [minChildSize] see [DraggableBottomSheet.minChildSize]
/// [maxChildSize] see [DraggableBottomSheet.maxChildSize]
/// [minFocusableExtent] see [DraggableBottomSheet.minFocusableExtent]
/// [updateInsets] see [DraggableBottomSheet.updateInsets]
/// [backgroundColor] the color of the [Material] wrapping this BottomSheet content which fills statusBar background
/// [elevation] of the BottomSheet. Defaults from [BottomSheetThemeData] or 0.
/// [shape] of the BottomSheet (will be scaled when extent matches statusBar). Defaults from [BottomSheetThemeData].
/// [clipBehavior] of the BottomSheet. Defaults from [BottomSheetThemeData] or [Clip.none].
/// [child] the actual content of the BottomSheet.
class FullscreenDraggableBottomSheet extends StatelessWidget {
  final double initialChildSize;
  final double minChildSize;
  final double maxChildSize;
  final double minFocusableExtent;
  final bool updateInsets;
  final Color? backgroundColor;
  final double? elevation;
  final ShapeBorder? shape;
  final Clip? clipBehavior;
  final Widget child;

  const FullscreenDraggableBottomSheet({
    Key? key,
    this.initialChildSize = .6,
    this.minChildSize = .2,
    this.maxChildSize = 1,
    this.minFocusableExtent = .5,
    this.updateInsets = true,
    this.backgroundColor,
    this.elevation,
    this.shape,
    this.clipBehavior,
    required this.child,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    final data = RootMediaQuery.of(context);
    final sheetTheme = Theme.of(context).bottomSheetTheme;
    return DraggableBottomSheet(
        initialChildSize: initialChildSize,
        minChildSize: minChildSize,
        maxChildSize: maxChildSize,
        minFocusableExtent: minFocusableExtent,
        updateInsets: updateInsets,
        builder: (context, notification) {
          final offsetTop = data.size.height * (1 - notification.extent);
          final heightTop = data.padding.top;
          final marginTop = max(heightTop - offsetTop, .0);
          final scale = heightTop > 0 ? min(offsetTop - heightTop, heightTop) / heightTop : 1.0;
          return Material(
              color: backgroundColor ?? sheetTheme.backgroundColor,
              elevation: elevation ?? sheetTheme.elevation ?? 0,
              shape: (shape ?? sheetTheme.shape)?.scale(scale),
              clipBehavior: clipBehavior ?? sheetTheme.clipBehavior ?? Clip.none,
              child: SizedBox(
                  height: data.size.height, child: Padding(padding: EdgeInsets.only(top: marginTop), child: child)));
        });
  }
}
