import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

/// Container that allows to build a Widget from it's states
/// [states] optional [MaterialStates] to control states from outside this widget
/// [onPressed] callback when widget is pressed
/// [onLongPress] callback when widget is long pressed
/// [minPressTime] when you press a widget tapping on it the state changes almost immediately. This can result
/// in a non fluid animation. If this value is greater than [Duration.zero] the pressed state is forced to remain
/// active for given duration
/// [drawSplash] if true this widget will draw [InkWell] splash
/// [builder] build widget from states
/// [child] passed back to builder
class MaterialStatefulWidget extends StatefulWidget {
  final MaterialStates? states;
  final VoidCallback? onPressed;
  final VoidCallback? onLongPress;
  final Duration minPressTime;
  final bool drawSplash;
  final ValueWidgetBuilder<MaterialStates> builder;
  final Widget? child;

  const MaterialStatefulWidget({
    Key? key,
    this.states,
    this.onPressed,
    this.onLongPress,
    this.minPressTime = const Duration(milliseconds: 100),
    this.drawSplash = false,
    required this.builder,
    this.child,
  }) : super(key: key);

  @override
  State<MaterialStatefulWidget> createState() => _MaterialStatefulWidgetState();
}

class _MaterialStatefulWidgetState extends State<MaterialStatefulWidget> {
  late MaterialStates _states;
  DateTime _pressedTime = DateTime.now();

  @override
  void initState() {
    super.initState();
    _states = widget.states ?? MaterialStates();
    _states._observers.add(_onStatesChanged);
  }

  @override
  void didUpdateWidget(MaterialStatefulWidget oldWidget) {
    super.didUpdateWidget(oldWidget);
    final newStates = widget.states;
    if (newStates != null && _states != newStates) {
      _states._observers.remove(_onStatesChanged);
      _states = newStates;
      _states._observers.add(_onStatesChanged);
    }
  }

  @override
  void dispose() {
    _states._observers.remove(_onStatesChanged);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) => InkWell(
      canRequestFocus: !_states.disabled,
      onFocusChange: (value) => _states.focused = value,
      onHighlightChanged: _onHighlightChanged,
      onHover: (value) => _states.hovered = value,
      splashFactory: widget.drawSplash ? null : NoSplash.splashFactory,
      highlightColor: widget.drawSplash ? null : Colors.transparent,
      onTap: widget.onPressed,
      onLongPress: widget.onLongPress,
      child: widget.builder(context, _states, widget.child));

  void _onHighlightChanged(value) {
    if (value) {
      _pressedTime = DateTime.now();
      _states.pressed = true;
      return;
    } else {
      final remaining = widget.minPressTime - DateTime.now().difference(_pressedTime);
      if (remaining > Duration.zero) {
        final time = _pressedTime;
        Future.delayed(remaining, () => time == _pressedTime ? _states.pressed = false : null);
      } else {
        _states.pressed = false;
      }
    }
  }

  void _onStatesChanged() => setState(() {});
}

/// Controller for [MaterialState] set used by [MaterialStatefulWidget]
/// You can update states by [update] function
@immutable
class MaterialStates {
  final Set<MaterialState> _states = {};
  final Set<VoidCallback> _observers = {};

  MaterialStates();

  bool get hovered => _states.contains(MaterialState.hovered);

  set hovered(bool value) => update(MaterialState.hovered, value: value);

  bool get focused => _states.contains(MaterialState.focused);

  set focused(bool value) => update(MaterialState.focused, value: value);

  bool get pressed => _states.contains(MaterialState.pressed);

  set pressed(bool value) => update(MaterialState.pressed, value: value);

  bool get dragged => _states.contains(MaterialState.dragged);

  set dragged(bool value) => update(MaterialState.dragged, value: value);

  bool get selected => _states.contains(MaterialState.selected);

  set selected(bool value) => update(MaterialState.selected, value: value);

  bool get disabled => _states.contains(MaterialState.disabled);

  set disabled(bool value) => update(MaterialState.disabled, value: value);

  bool get error => _states.contains(MaterialState.error);

  set error(bool value) => update(MaterialState.error, value: value);

  bool update(MaterialState state, {required bool value}) {
    final result = value ? _states.add(state) : _states.remove(state);
    if (result) for (final notify in _observers) notify();
    return result;
  }

  @override
  bool operator ==(Object other) => other is MaterialStates && setEquals(other._states, _states);

  @override
  int get hashCode => hashValues(runtimeType, hashList(_states));

  @override
  String toString() => 'MaterialStates[${_states.join(', ')}]';
}
